import json
from datetime import datetime
from typing import Literal, TYPE_CHECKING

if TYPE_CHECKING:
    from .chains import Chain
    from .explorer import Explorer


def fetch_abi(explorer: "Explorer", address: str):
    contract_info = explorer.contract.getsourcecode(address=address)[0]
    abi = json.loads(contract_info["ABI"])
    if contract_info["Proxy"] == "1":
        implementation_info = explorer.contract.getsourcecode(
            address=contract_info["Implementation"]
        )[0]
        abi = json.loads(implementation_info["ABI"])

    return abi


def get_block_by_datetime(
    explorer: "Explorer", dt: datetime, closest: Literal["before", "after"] = "before"
):
    return explorer.block.getblocknobytime(
        timestamp=int(dt.timestamp()), closest=closest
    )


def get_datetime_by_block(chain: "Chain", block: int):
    return datetime.fromtimestamp(chain.eth.get_block(block).timestamp)



import time
import requests
from typing import Callable


class ExplorerError(RuntimeError):
    def __init__(self, message, data) -> None:
        super().__init__(message)

        self.data = data

    def __str__(self) -> str:
        return f"{super().__str__()}.\nReturned result: {self.data}"


class ExplorerAction:
    def __init__(self, module: "ExplorerModule", action: str) -> None:
        self.module = module
        self.action = action

    def __call__(self, **params):
        params["module"] = str(self.module)
        params["action"] = self.action
        if self.module.explorer.api_key is not None:
            params["apikey"] = self.module.explorer.api_key

        if "apikey" not in params:
            # stupid throtling
            time.sleep(5)

        r = requests.get(self.module.explorer.url, params=params)
        result = r.json()
        if result["status"] != "1":
            raise ExplorerError(result["message"], result["result"])
        return result["result"]


class ExplorerModule:
    def __init__(self, explorer: "Explorer", module: str) -> None:
        self.explorer = explorer
        self.name = module

    def __str__(self) -> str:
        return self.name

    def __getattr__(self, action: str) -> Callable:
        return ExplorerAction(self, action)


class Explorer:
    def __init__(self, url, api_key=None) -> None:
        self.url = url
        self.api_key = None

    def set_api_key(self, api_key):
        self.api_key = api_key

    def __getattr__(self, module: str) -> ExplorerModule:
        return ExplorerModule(self, module)


etherscan = Explorer("https://api.etherscan.io/api")
snowtrace = Explorer("https://api.snowtrace.io/api")
ftmscan = Explorer("https://api.ftmscan.com/api")
arbiscan = Explorer("https://api.arbiscan.io/api")
bscscan = Explorer("https://api.bscscan.com/api")
optimistic_etherscan = Explorer("https://api-optimistic.etherscan.io/api")
polygonscan = Explorer("https://api.polygonscan.com/api")



from dataclasses import dataclass
from typing import Optional

from web3 import Web3
from web3.middleware import geth_poa_middleware

from .contract import Contract
from .explorer import (
    Explorer,
    etherscan,
    snowtrace,
    ftmscan,
    arbiscan,
    bscscan,
    optimistic_etherscan,
    polygonscan,
)
from . import utils


@dataclass
class NativeToken:
    symbol: str
    decimals: int = 18


class Chain(Web3):
    explorer: Optional[Explorer]
    native_token: Optional[NativeToken]

    def contract(chain):
        def _contract(*args, **kwargs):
            return Contract(*args, chain=chain, **kwargs)

        return _contract

    def get_datetime_by_block(chain):
        def _get_datetime_by_block(block):
            return utils.get_datetime_by_block(chain, block)

        return _get_datetime_by_block


def chain_w3(
    rpc, explorer=None, symbol: str = None, decimals=18, poa_middleware=True
) -> Chain:
    w3 = Web3(Web3.HTTPProvider(rpc))
    w3.explorer = explorer
    w3.native_token = NativeToken(symbol, decimals) if symbol else None
    if poa_middleware:
        w3.middleware_onion.inject(geth_poa_middleware, layer=0)

    w3.contract = Chain.contract(w3)
    w3.get_datetime_by_block = Chain.get_datetime_by_block(w3)

    return w3


ethereum = chain_w3("https://rpc.ankr.com/eth", etherscan, "ETH")
avalanche = chain_w3("https://rpc.ankr.com/avalanche", snowtrace, "AVAX")
fantom = chain_w3("https://rpc.ankr.com/fantom", ftmscan, "FTM")
arbitrum = chain_w3("https://rpc.ankr.com/arbitrum", arbiscan, "ETH")
bsc = chain_w3("https://rpc.ankr.com/bsc", bscscan, "BNB")
optimism = chain_w3("https://rpc.ankr.com/optimism", optimistic_etherscan, "ETH")
polygon = chain_w3("https://rpc.ankr.com/polygon", polygonscan, "MATIC")



from typing import TYPE_CHECKING

from eth_utils import is_hex_address, to_checksum_address


if TYPE_CHECKING:
    from .chains import Chain

from .explorer import Explorer
from .utils import fetch_abi


class Contract:
    def __init__(
        self, address, chain: "Chain", explorer: Explorer = None, abi=None
    ) -> None:
        self.address = to_checksum_address(address)
        self.explorer = explorer or chain.explorer
        self.abi = abi or fetch_abi(self.explorer, address)
        self._contract = chain.eth.contract(self.address, abi=self.abi)

    def __getattr__(self, method_name):
        def fn(*args, block="latest", **kwargs):
            args = [to_checksum_address(a) if is_hex_address(a) else a for a in args]
            kwargs = {
                k: to_checksum_address(v) if is_hex_address(v) else v
                for k, v in kwargs.items()
            }

            return self._contract.functions[method_name](*args, **kwargs).call(
                block_identifier=block
            )

        return fn



import datetime
import time

# import pandas as pd
# from tqdm import trange

from src.web3_premium.explorer import snowtrace
from src.web3_premium.contract import Contract
from src.web3_premium.chains import avalanche


snowtrace.set_api_key("EW1TBC2J8P6TNGARYFE4U1895DBW6MQ354")
Contract("0x486Af39519B4Dc9a7fCcd318217352830E8AD9b4", avalanche, snowtrace)


import src.web3_premium.utils as utils

print(utils.get_block_by_datetime(snowtrace, datetime.datetime.now()))
